// ======================================================================
//
// geometry.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Geometry.h"

#include "GL_abstract.h"
#include "light.h"

// ======================================================================
// these variables are not static members of the class to they can be
// referenced in inline assembly

static real             viewportCenterX;
static real             viewportCenterY;
static real             viewportWidthMultiplier;
static real             viewportHeightMultiplier;
 
// ======================================================================

bool                    Geometry::lightingEnabled;
bool                    Geometry::needObjectToCameraRebuilt;
bool                    Geometry::needLightingRecalculated;

real                    Geometry::nearPlane;
real                    Geometry::farPlane;

real                    Geometry::clipWidthMultiplier;
real                    Geometry::clipHeightMultiplier;

Vector                  Geometry::frustumNormal[6];
real                    Geometry::frustumPlaneD[6];

Xform                   Geometry::objectToWorld;
Xform                   Geometry::worldToCamera;
Xform                   Geometry::objectToCamera;

int                     Geometry::lightCount;
Light                 **Geometry::lightList;
int                     Geometry::applyLightCount;
int                     Geometry::applyLightSize;
Light                 **Geometry::applyLightList;

const Vector           *Geometry::vertex;
int                     Geometry::vertexCount;
const Vector           *Geometry::normal;
int                     Geometry::normalCount;

int                     Geometry::viewspaceVertexSize;
Vector                 *Geometry::viewspaceVertex;
int8                   *Geometry::viewspaceVertexClipFlags;

TVertex                 Geometry::transformedVertexBuffer[TVERTEX_BUFFER];
TVertex                 Geometry::clippedVertexBuffer[CLIPPED_BUFFER];

GL_abstract            *Geometry::gl;

// ======================================================================

void Geometry::install(GL_abstract *newGl)
{
	lightingEnabled           = true;

	gl                        = newGl;

	viewspaceVertexSize       = USE_GEOMETRY_VERTEX_ARRAY_INITIAL_SIZE;
	viewspaceVertex           = NEW Vector[viewspaceVertexSize];
	viewspaceVertexClipFlags  = NEW int8[viewspaceVertexSize];

	applyLightSize            = USE_GEOMETRY_LIGHT_ARRAY_INITIAL_SIZE;
	applyLightList            = NEW Light*[applyLightSize];
}

// ----------------------------------------------------------------------

void Geometry::remove(void)
{
	delete [] viewspaceVertex;
	viewspaceVertex= NULL;

	delete [] viewspaceVertexClipFlags;
	viewspaceVertexClipFlags = NULL;

	delete [] applyLightList;
	applyLightList = NULL;
}

// ----------------------------------------------------------------------

void Geometry::setLightingEnabled(bool newLightingEnabled)
{
	lightingEnabled = newLightingEnabled;
}

// ----------------------------------------------------------------------

void Geometry::setLightList(Light **newLightList, int newLightCount)
{
	lightList = newLightList;
	lightCount = newLightCount;
	needLightingRecalculated = true;

	if (lightCount > applyLightSize)
	{
		DEBUG_FATAL(USE_GEOMETRY_LIGHT_ARRAY_GROW_SIZE == 0, ("Geometry::setLightList light overflow"));

		while (applyLightSize < lightCount)
			applyLightSize += USE_GEOMETRY_LIGHT_ARRAY_GROW_SIZE;

		delete [] applyLightList;
		applyLightList = NEW Light*[applyLightSize];

		WARNING(true, ("Geometry::setLightList number of lights grown to %d\n", applyLightSize));
	}
}

// ----------------------------------------------------------------------

void Geometry::setCamera(const Camera &camera)
{
	int i;

	needObjectToCameraRebuilt = true;
	needLightingRecalculated  = true;
	worldToCamera.invert(camera.getObjectToWorldXform());

	nearPlane                = camera.getNearPlane();
	farPlane                 = camera.getFarPlane();
	viewportCenterX          = real(camera.getCenterX());
	viewportCenterY          = real(camera.getCenterY());
	clipWidthMultiplier      = camera.getClipWidthMultiplier();
	clipHeightMultiplier     = camera.getClipHeightMultiplier();
	viewportWidthMultiplier  = camera.getWidthMultiplier();
	viewportHeightMultiplier = camera.getHeightMultiplier();

	for (i = 0; i < 6; ++i)
	{
		frustumNormal[i] = camera.getFrustumNormal(i);
		frustumPlaneD[i] = camera.getFrustumPlaneD(i);
	}
}

// ----------------------------------------------------------------------

void Geometry::rebuildTransformsReal(void)
{
	++Camera::renderStats->matrixMultiplies;
	objectToCamera.multiply(worldToCamera, objectToWorld);
	needObjectToCameraRebuilt = false;
}

// ----------------------------------------------------------------------

void Geometry::recalculateLightingReal(void)
{
	int i;

	for (i = 0, applyLightCount = 0; i < lightCount; ++i)
		if (lightList[i]->setup(objectToWorld, NULL))
			applyLightList[applyLightCount++] = lightList[i];

	needLightingRecalculated = false;
}

// ----------------------------------------------------------------------

bool Geometry::setVertexList(const Vector *newVertex, int newVertexCount, const Vector *newNormal, int newNormalCount)
{
	rebuildTransforms();

	needLightingRecalculated = true;

	vertex      = newVertex;
	vertexCount = newVertexCount;
	normal      = newNormal;
	normalCount = newNormalCount;

	if (vertexCount >= viewspaceVertexSize)
	{
		DEBUG_FATAL(USE_GEOMETRY_VERTEX_ARRAY_GROW_SIZE == 0, ("Geometry::setVertexList vertex overflow"));

		delete [] viewspaceVertex;
		viewspaceVertex = NULL;

		delete [] viewspaceVertexClipFlags;
		viewspaceVertexClipFlags = NULL;

		while (viewspaceVertexSize < vertexCount)
			viewspaceVertexSize += USE_GEOMETRY_VERTEX_ARRAY_GROW_SIZE;

		// grow the vertex array
		viewspaceVertex          = NEW Vector[viewspaceVertexSize];
		viewspaceVertexClipFlags = NEW int8[viewspaceVertexSize];

		WARNING(true, ("Geometry::setVertexList number of vertices per object grown to %d\n", viewspaceVertexSize));
	}

	Camera::renderStats->verticesTransformed += vertexCount;

	// make one pass and transform all vertices into view space
	if (!objectToCamera.rotateTranslatePerspectiveClip(vertex, viewspaceVertex, vertexCount, nearPlane, farPlane, clipWidthMultiplier, clipHeightMultiplier, viewspaceVertexClipFlags))
		return false;

	return true;
}

// ----------------------------------------------------------------------

void Geometry::drawPointList(const UVertex *vertexList, int numberOfVertices)
{
	int    out;
	real   ooz;

	for (out = 0; numberOfVertices; --numberOfVertices, ++vertexList)
		if (viewspaceVertexClipFlags[vertexList->vertexIndex] == 0)
		{
			const Vector &v  = viewspaceVertex[vertexList->vertexIndex];
			TVertex      &tv = transformedVertexBuffer[out];

			// project point
			ooz = RECIP(v.z);
			tv.x   = (v.x * ooz * viewportWidthMultiplier) + viewportCenterX;
			tv.y   = (v.y * ooz * viewportHeightMultiplier) + viewportCenterY;
			tv.z   = v.z;
			tv.ooz = RECIP(v.z);

			// copy uv info
			tv.u   = vertexList->u;
			tv.v   = vertexList->v;

			// copy lighting info
			tv.r   = vertexList->r;
			tv.g   = vertexList->g;
			tv.b   = vertexList->b;
			tv.a   = vertexList->a;

			// another output vertex
			if (++out == TVERTEX_BUFFER_SAFE)
			{
				gl->submitPointList(transformedVertexBuffer, out);
				out = 0;
			}
		}

	if (out)
		gl->submitPointList(transformedVertexBuffer, out);
}

// ----------------------------------------------------------------------

void Geometry::drawLineList(const UVertex *vertexList, int numberOfVertices)
{
	int    out, flag0, flag1;
	real   ooz;

	for (out = 0; numberOfVertices; numberOfVertices -= 2, vertexList += 2)
	{
		// get the clipping flags
		flag0 = viewspaceVertexClipFlags[vertexList[0].vertexIndex];
		flag1 = viewspaceVertexClipFlags[vertexList[1].vertexIndex];

		// test the whole line for trivial rejection
		if ((flag0 & flag1) != 0)
			continue;

		// setup the output vertices
		TVertex      &tv0 = transformedVertexBuffer[out];
		TVertex      &tv1 = transformedVertexBuffer[out+1];

		// copy v0
		tv0.x   = viewspaceVertex[vertexList[0].vertexIndex].x;
		tv0.y   = viewspaceVertex[vertexList[0].vertexIndex].y;
		tv0.z   = viewspaceVertex[vertexList[0].vertexIndex].z;
		tv0.u   = vertexList[0].u;
		tv0.v   = vertexList[0].v;
		tv0.r   = vertexList[0].r;
		tv0.g   = vertexList[0].g;
		tv0.b   = vertexList[0].b;
		tv0.a   = vertexList[0].a;

		// copy v1
		tv1.x   = viewspaceVertex[vertexList[1].vertexIndex].x;
		tv1.y   = viewspaceVertex[vertexList[1].vertexIndex].y;
		tv1.z   = viewspaceVertex[vertexList[1].vertexIndex].z;
		tv1.u   = vertexList[1].u;
		tv1.v   = vertexList[1].v;
		tv1.r   = vertexList[1].r;
		tv1.g   = vertexList[1].g;
		tv1.b   = vertexList[1].b;
		tv1.a   = vertexList[1].a;

		// line segment that needs clipping
		if ((flag0 | flag1) != 0)
		{
			int i;
			bool skip;

			for (skip = false, i = 0 ; i < 6; ++i)
			{
				real side0 = frustumNormal[i].dot(Vector(tv0.x, tv0.y, tv0.z)) + frustumPlaneD[i];
				real side1 = frustumNormal[i].dot(Vector(tv1.x, tv1.y, tv1.z)) + frustumPlaneD[i];

				if (side0 > CONST_REAL(0) && side1 > CONST_REAL(0))
				{
					skip = true;
					break;
				}

				if (opposingSignsNotZero(side0, side1))
				{
					real t = abs(side0) / (abs(side0) + abs(side1));

					if (side0 > CONST_REAL(0))
					{
						tv0.x = tv0.x + (tv1.x - tv0.x) * t;
						tv0.y = tv0.y + (tv1.y - tv0.y) * t;
						tv0.z = tv0.z + (tv1.z - tv0.z) * t;
						tv0.u = tv0.u + (tv1.u - tv0.u) * t;
						tv0.v = tv0.v + (tv1.v - tv0.v) * t;
						tv0.r = tv0.r + (tv1.r - tv0.r) * t;
						tv0.g = tv0.g + (tv1.g - tv0.g) * t;
						tv0.b = tv0.b + (tv1.b - tv0.b) * t;
						tv0.a = tv0.a + (tv1.a - tv0.a) * t;
					}
					else
					{
						tv1.x = tv0.x + (tv1.x - tv0.x) * t;
						tv1.y = tv0.y + (tv1.y - tv0.y) * t;
						tv1.z = tv0.z + (tv1.z - tv0.z) * t;
						tv1.u = tv0.u + (tv1.u - tv0.u) * t;
						tv1.v = tv0.v + (tv1.v - tv0.v) * t;
						tv1.r = tv0.r + (tv1.r - tv0.r) * t;
						tv1.g = tv0.g + (tv1.g - tv0.g) * t;
						tv1.b = tv0.b + (tv1.b - tv0.b) * t;
						tv1.a = tv0.a + (tv1.a - tv0.a) * t;
					}
				}
			}

			if (skip)
				continue;
		}

		// project first vertex
		ooz = RECIP(tv0.z);
		tv0.x   = (tv0.x * ooz * viewportWidthMultiplier) + viewportCenterX;
		tv0.y   = (tv0.y * ooz * viewportHeightMultiplier) + viewportCenterY;
		tv0.ooz = ooz;

		// project second vertex
		ooz = RECIP(tv1.z);
		tv1.x   = (tv1.x * ooz * viewportWidthMultiplier) + viewportCenterX;
		tv1.y   = (tv1.y * ooz * viewportHeightMultiplier) + viewportCenterY;
		tv1.ooz = ooz;

		// update the number of output vertices
		out += 2;

		// output line segment
		if (out >= TVERTEX_BUFFER_SAFE)
		{
			gl->submitLineList(transformedVertexBuffer, out);
			out = 0;
		}
	}

	if (out)
		gl->submitLineList(transformedVertexBuffer, out);
}

// ----------------------------------------------------------------------
// -qq- optimize this

void Geometry::drawLineStrip(const UVertex *vertexList, int numberOfVertices)
{
	int     i;
	UVertex v[2];

	for (i = 0; i < numberOfVertices-1; ++i)
	{
		v[0] = vertexList[i+0];
		v[1] = vertexList[i+1];
		drawLineList(v, 2);
	}
}

// ----------------------------------------------------------------------
// -qq- OPTIMIZE this

void Geometry::drawTriangleList(const UVertex *vertexList, int numberOfVertices)
{
	int          out, flag0, flag1, flag2, i, j, sourceCount, clippedCount;
	real         ooz;
	TVertex     *source, *clipped;
	ColorVector  total, component;

	// setup 
	if (lightingEnabled)
		recalculateLighting();
	else
		total.setRGBM(CONST_REAL(1.0));

	for (out = 0; numberOfVertices; numberOfVertices -= 3, vertexList += 3)
	{
		flag0 = viewspaceVertexClipFlags[vertexList[0].vertexIndex];
		flag1 = viewspaceVertexClipFlags[vertexList[1].vertexIndex];
		flag2 = viewspaceVertexClipFlags[vertexList[2].vertexIndex];

		// test the triangle for trivial rejection
		if ((flag0 & flag1 & flag2) != 0)
			continue;

		// setup the vertices
		for (i = 0; i < 3; ++i)
		{
			const Vector &v  = viewspaceVertex[vertexList[i].vertexIndex];
			TVertex      &tv = transformedVertexBuffer[out+i];

			// compute lighting for this vertex
			if (lightingEnabled)
			{
				total.makeZero();
				
				for (j = 0; j < applyLightCount; ++j)
				{
					applyLightList[j]->apply(objectToWorld, vertex[vertexList[i].vertexIndex], normal[vertexList[i].normalIndex], &component);
					total += component;
				}

				total.clampZeroToOne();
			}

			// copy vertex data
			tv.x   = v.x;
			tv.y   = v.y;
			tv.z   = v.z;
			tv.u   = vertexList[i].u;
			tv.v   = vertexList[i].v;

			// setup light
			tv.r   = vertexList[i].r * total.getR();
			tv.g   = vertexList[i].g * total.getG();
			tv.b   = vertexList[i].b * total.getB();
			tv.m   = vertexList[i].m * total.getM();
			tv.a   = vertexList[i].a;
		}

		// setup the source triangle
		sourceCount  = 3;
		source       = transformedVertexBuffer;

		// setup the dest vertex
		clipped      = clippedVertexBuffer;
		clippedCount = 0;

		// clip polygon
		if ((flag0 | flag1 | flag2) != 0)
		{
			// clip against all clipping planes
			for (i = 0; i < 6 && sourceCount > 2; ++i)
			{
				int v0, v1;

				// clip all edges of the incoming polygon
				for (v1 = 0, v0 = sourceCount-1; v1 < sourceCount; v0 = v1, ++v1)
				{
					// -qq- OPTIMIZE obviously don't need to recalculate both each time
					real side0 = frustumNormal[i].dot(source[v0].x, source[v0].y, source[v0].z) + frustumPlaneD[i];
					real side1 = frustumNormal[i].dot(source[v1].x, source[v1].y, source[v1].z) + frustumPlaneD[i];

					// see if the line segment crosses the viewing frustum plane
					if (opposingSignsNotZero(side0, side1))
					{
						real t = abs(side0) / (abs(side0) + abs(side1));

						// use parametric equations to clip the line segment
						clipped[clippedCount].x = source[v0].x + (source[v1].x - source[v0].x) * t;
						clipped[clippedCount].y = source[v0].y + (source[v1].y - source[v0].y) * t;
						clipped[clippedCount].z = source[v0].z + (source[v1].z - source[v0].z) * t;
						clipped[clippedCount].u = source[v0].u + (source[v1].u - source[v0].u) * t;
						clipped[clippedCount].v = source[v0].v + (source[v1].v - source[v0].v) * t;
						clipped[clippedCount].r = source[v0].r + (source[v1].r - source[v0].r) * t;
						clipped[clippedCount].g = source[v0].g + (source[v1].g - source[v0].g) * t;
						clipped[clippedCount].b = source[v0].b + (source[v1].b - source[v0].b) * t;
						clipped[clippedCount].a = source[v0].a + (source[v1].a - source[v0].a) * t;

						++clippedCount;
					}

					// see if the destination vertex is within the view frustum
					if (side1 <= CONST_REAL(0))
					{
						clipped[clippedCount] = source[v1];
						++clippedCount;
					}
				}

				// swap the working vertex buffers
				swap(source, clipped);
				sourceCount = clippedCount;
				clippedCount = 0;
			}
		}

		// see if we still have a valid polygon
		if (sourceCount > 2)
		{
			// project all the vertices and map them into the viewport
			for (i = 0; i < sourceCount; ++i)
			{
				TVertex &tv = source[out+i];

				ooz    = RECIP(tv.z);
				tv.x   = (tv.x * ooz * viewportWidthMultiplier) + viewportCenterX;
				tv.y   = (tv.y * ooz * viewportHeightMultiplier) + viewportCenterY;
				tv.ooz = ooz;
			}

			// copy the vertices into the clipped area if we have to build a fan
			if (source == transformedVertexBuffer)
			{
				for (i = 0; i < sourceCount; ++i)
					clipped[i] = source[i];
				source = clipped;
			}

			// add all the triangles in the fan
			for (i = 0; i < sourceCount-2; ++i)
			{
				// backface cull
				real area =
					(source[0  ].x * source[i+1].y) - (source[0  ].y * source[i+1].x) +
					(source[i+1].x * source[i+2].y) - (source[i+1].y * source[i+2].x) +
					(source[i+2].x * source[0  ].y) - (source[i+2].y * source[  0].x);

				if (area < 0)
					continue;

				// update the number of pixels rendered
				++Camera::renderStats->polygonRenderArea += (REAL_TO_INT(area) / 2);

				// copy the output vertices
				transformedVertexBuffer[out+0] = source[0  ];
				transformedVertexBuffer[out+1] = source[i+1];
				transformedVertexBuffer[out+2] = source[i+2];
				out += 3;

				// submit if the vertex list is getting full
				if (out >= TVERTEX_BUFFER_SAFE)
				{
					gl->submitTriangleList(transformedVertexBuffer, out);
					out = 0;
				}
			}
		}
	}

	if (out)
		gl->submitTriangleList(transformedVertexBuffer, out);
}

// ----------------------------------------------------------------------
// -qq- optimize this

void Geometry::drawTriangleStrip(const UVertex *vertexList, int numberOfVertices)
{
	int     i;
	UVertex v[3];

	for (i = 0; i < numberOfVertices-2; ++i)
	{
		// reverse every other triangle to keep them clockwise
		if ((i & 1) == 0)
		{
			v[0] = vertexList[i+0];
			v[1] = vertexList[i+1];
			v[2] = vertexList[i+2];
		}
		else
		{
			v[0] = vertexList[i+0];
			v[1] = vertexList[i+2];
			v[2] = vertexList[i+1];
		}

		drawTriangleList(v, 3);
	}
}

// ----------------------------------------------------------------------
// -qq- optimize this

void Geometry::drawTriangleFan(const UVertex *vertexList, int numberOfVertices)
{
	int     i;
	UVertex v[3];

	// setup the first vertex
	v[0] = vertexList[0];

	// render all the slices in the fan
	for (i = 0; i < numberOfVertices-2; ++i)
	{
		v[1] = vertexList[i+1];
		v[2] = vertexList[i+2];
		drawTriangleList(v, 3);
	}
}

// ======================================================================
